﻿/*
 * Customer    :  Dot.Blog
 * Project     :  VirtualInCtor
 * Description :  Contourner le problème de l'appel d'un membre virtuel dans le constructeur
 * Version     :  1.0
 * Modified on :  mercredi 5 novembre 2008 16:14
 *
 * Copyright 2008 Olivier Dahan - www.e-naxos.com/blog
 *
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace VirtualInCtor
{

    #region Exemple 1 : mauvaise implémentation

    // Classe Parent
    public class Parent
    {
        public Parent(int valeur)
        {
            MethodeVirtuelle();
        }
        public virtual void MethodeVirtuelle()
        { }
    }
    
    // Classe dérivée
    public class Enfant : Parent
    {
        private int val;

        public Enfant(int valeur) : base(valeur)
        {
            val = valeur;
        }

        public override void MethodeVirtuelle()
        {
            Console.WriteLine("Classe Enfant. champ val = "+val);
            // écrira toujours "0"
        }
    }

    #endregion


    #region Exemple 2 : Avec init séparé

    // Classe Parent
    public class Parent2
    {
        public Parent2(int valeur)
        {
            // MethodeVirtuelle();
        }

        public virtual void Init()
        {
            MethodeVirtuelle();
        }

        public virtual void MethodeVirtuelle()
        { }
    }

    // Classe dérivée
    public class Enfant2 : Parent2
    {
        private int val;

        public Enfant2(int valeur)
            : base(valeur)
        {
            val = valeur;
        }

        public override void MethodeVirtuelle()
        {
            Console.WriteLine("Classe Enfant2. champ val = " + val);
        }
    }

    #endregion


    #region Exemple 3 : Avec Factory

    // Classe Parent
    public class Parent3
    {
        public Parent3(int valeur)
        {
            // MethodeVirtuelle();
        }

        public virtual void Init()
        {
            MethodeVirtuelle();
        }

        public virtual void MethodeVirtuelle()
        { }
    }

    // Classe dérivée
    public class Enfant3 : Parent3
    {
        private int val;

        public static Enfant3 CreerInstance(int valeur)
        {
            var enfant3 = new Enfant3(valeur);
            enfant3.Init();
            return enfant3;
        }

        protected Enfant3(int valeur)
            : base(valeur)
        {
            val = valeur;
        }

        public override void MethodeVirtuelle()
        {
            Console.WriteLine("Classe Enfant3. champ val = " + val);
        }
    }

    #endregion

    class Program
    {
        static void Main(string[] args)
        {
            // Méthode 1 : mauvaise implémentation
            // affiche 0 car la classe dérivée n'est pas encore initialisée
            // lorsque la méthode virtuelle est appelée depuis le constructeur de
            // la classe Parent
            var enfant = new Enfant(10);

            // Méthode 2 : avec init séparé
            var enfant2 = new Enfant2(10);
            enfant2.Init(); // affichera 10

            // Méthode 3 avec Factory
            // var enfant3 = new Enfant3(); - Impossible car Ctor protégé
            var enfant3 = Enfant3.CreerInstance(10);
            // affiche 10, sans obligé l'utilisateur de la classe à l'appel de Init()


            Console.ReadLine();

        }
    }
}
